import json
from datetime import datetime, timezone
from typing import ClassVar, Optional, cast

import neo4j
from neo4j import AsyncDriver, AsyncGraphDatabase
from neo4j.exceptions import ResultNotSingleError
from pydantic import Field

from icij_worker import AsyncApp, Task, TaskState
from icij_worker.constants import (
    NEO4J_TASK_CANCELLED_BY_EVENT_REL,
    NEO4J_TASK_CANCEL_EVENT_CANCELLED_AT,
    NEO4J_TASK_CANCEL_EVENT_EFFECTIVE,
    NEO4J_TASK_CANCEL_EVENT_NODE,
    NEO4J_TASK_CANCEL_EVENT_REQUEUE,
    NEO4J_TASK_ID,
    NEO4J_TASK_LOCK_NODE,
    NEO4J_TASK_LOCK_TASK_ID,
    NEO4J_TASK_LOCK_WORKER_ID,
    NEO4J_TASK_MANAGER_EVENT_EVENT,
    NEO4J_TASK_MANAGER_EVENT_NODE,
    NEO4J_TASK_MANAGER_EVENT_NODE_CREATED_AT,
    NEO4J_TASK_NODE,
    NEO4J_TASK_PROGRESS,
    NEO4J_TASK_RETRIES_LEFT,
    TASK_ID,
)
from icij_worker.exceptions import (
    MessageDeserializationError,
    TaskQueueIsFull,
    UnknownTask,
)
from icij_worker.objects import AsyncBackend, ManagerEvent, Message
from icij_worker.task_manager import TaskManager, TaskManagerConfig
from icij_worker.utils import FromConfig
from icij_worker.utils.neo4j_ import Neo4jConsumerMixin


@TaskManagerConfig.register()
class Neo4JTaskManagerConfig(TaskManagerConfig):
    backend: ClassVar[AsyncBackend] = Field(const=True, default=AsyncBackend.neo4j)
    event_refresh_interval_s: float = 0.1
    neo4j_host: str = "localhost"
    neo4j_port: int = 7687
    neo4j_user: str = "neo4j"
    neo4j_password: str = "theneo4jpassword"
    neo4j_scheme: str = "neo4j"

    def to_driver(self) -> AsyncDriver:
        uri = f"{self.neo4j_scheme}://{self.neo4j_host}:{self.neo4j_port}"
        auth = neo4j.basic_auth(self.neo4j_user, self.neo4j_password)
        driver = AsyncGraphDatabase.driver(uri, auth=auth)
        return driver


@TaskManager.register(AsyncBackend.neo4j)
class Neo4JTaskManager(TaskManager, Neo4jConsumerMixin):

    def __init__(
        self,
        app: AsyncApp,
        driver: neo4j.AsyncDriver,
        event_refresh_interval_s: float = 0.1,
    ) -> None:
        super().__init__(app)
        super(TaskManager, self).__init__(driver)
        self._event_refresh_interval_s = event_refresh_interval_s

    @classmethod
    def _from_config(cls, config: Neo4JTaskManagerConfig, **extras) -> FromConfig:
        driver = config.to_driver()
        tm = cls(
            config.app, driver, event_refresh_interval_s=config.event_refresh_interval_s
        )
        return tm

    @property
    def driver(self) -> neo4j.AsyncDriver:
        return self._driver

    async def _enqueue(self, task: Task):
        # pylint: disable=arguments-differ
        db = await self._get_task_db(task_id=task.id)
        async with self._db_session(db) as sess:
            await sess.execute_write(
                _enqueue_task_tx,
                task_id=task.id,
                max_queue_size=self.max_task_queue_size,
            )

    async def _requeue(self, task: Task):
        # pylint: disable=arguments-differ
        db = await self._get_task_db(task_id=task.id)
        async with self._db_session(db) as sess:
            return await sess.execute_write(
                _requeue_task_tx,
                task_id=task.id,
                retries_left=task.retries_left,
                max_queue_size=self.max_task_queue_size,
            )

    async def _consume(self) -> ManagerEvent:
        event_as_json = await self._consume_(
            _consume_manager_events_tx,
            refresh_interval_s=self._event_refresh_interval_s,
            db_filter=None,
        )
        try:
            event = Message.parse_obj(json.loads(event_as_json))
        except Exception as e:
            msg = f"invalid event {event_as_json}"
            raise MessageDeserializationError(msg) from e
        return cast(ManagerEvent, event)

    async def cancel(self, task_id: str, *, requeue: bool):
        async with self._task_session(task_id) as sess:
            await sess.execute_write(_cancel_task_tx, task_id=task_id, requeue=requeue)


async def _enqueue_task_tx(
    tx: neo4j.AsyncTransaction, *, task_id: str, max_queue_size: int
) -> Task:
    count_query = f"""MATCH (task:{NEO4J_TASK_NODE}:`{TaskState.QUEUED.value}`)
RETURN count(task.id) AS nQueued
"""
    res = await tx.run(count_query)
    count = await res.single(strict=True)
    n_queued = count["nQueued"]
    if max_queue_size is not None and n_queued >= max_queue_size:
        raise TaskQueueIsFull(max_queue_size)

    query = f"""MATCH (t:{NEO4J_TASK_NODE} {{ {NEO4J_TASK_ID}: $taskId }})
WITH t
CALL apoc.create.setLabels(t, $labels) YIELD node AS task
RETURN task
"""
    labels = [NEO4J_TASK_NODE, TaskState.QUEUED.value]
    res = await tx.run(query, taskId=task_id, labels=labels)
    recs = [rec async for rec in res]
    if not recs:
        raise UnknownTask(task_id)
    if len(recs) > 1:
        raise ValueError(f"Multiple tasks found for task {task_id}")
    return Task.from_neo4j(recs[0])


async def _cancel_task_tx(tx: neo4j.AsyncTransaction, task_id: str, requeue: bool):
    query = f"""MATCH (task:{NEO4J_TASK_NODE} {{ {NEO4J_TASK_ID}: $taskId }})
CREATE (task)-[
    :{NEO4J_TASK_CANCELLED_BY_EVENT_REL}
]->(:{NEO4J_TASK_CANCEL_EVENT_NODE} {{ 
        {NEO4J_TASK_CANCEL_EVENT_CANCELLED_AT}: $cancelledAt, 
        {NEO4J_TASK_CANCEL_EVENT_EFFECTIVE}: false,
        {NEO4J_TASK_CANCEL_EVENT_REQUEUE}: $requeue
    }})
"""
    await tx.run(
        query, taskId=task_id, requeue=requeue, cancelledAt=datetime.now(timezone.utc)
    )


async def _consume_manager_events_tx(tx: neo4j.AsyncTransaction) -> Optional[str]:
    get_event_query = f"""MATCH (event:{NEO4J_TASK_MANAGER_EVENT_NODE})
RETURN event.{NEO4J_TASK_MANAGER_EVENT_EVENT} as eventAsJson
ORDER BY event.{NEO4J_TASK_MANAGER_EVENT_NODE_CREATED_AT} ASC
LIMIT 1
"""
    res = await tx.run(get_event_query)
    try:
        event = await res.single(strict=True)
    except ResultNotSingleError:
        return None
    return event["eventAsJson"]


async def _requeue_task_tx(
    tx: neo4j.AsyncTransaction, *, task_id: str, retries_left: int, max_queue_size: int
):
    count_query = f"""MATCH (task:{NEO4J_TASK_NODE}:`{TaskState.QUEUED.value}`)
RETURN count(task.id) AS nQueued
"""
    res = await tx.run(count_query)
    count = await res.single(strict=True)
    n_queued = count["nQueued"]
    if max_queue_size is not None and n_queued >= max_queue_size:
        raise TaskQueueIsFull(max_queue_size)
    query = f"""MATCH (t:{NEO4J_TASK_NODE} {{ {TASK_ID}: $taskId}})
SET t.{NEO4J_TASK_RETRIES_LEFT} = $retriesLeft, t.{NEO4J_TASK_PROGRESS} = 0.0
WITH t
CALL apoc.create.setLabels(t, $labels) YIELD node AS task
RETURN task
"""
    labels = [NEO4J_TASK_NODE, TaskState.QUEUED.value]
    res = await tx.run(query, taskId=task_id, retriesLeft=retries_left, labels=labels)
    try:
        await res.single(strict=True)
    except ResultNotSingleError as e:
        raise UnknownTask(task_id) from e


async def _dlq_task_tx(tx: neo4j.AsyncTransaction, *, task_id: str, worker_id: str):
    query = f"""MATCH (lock:{NEO4J_TASK_LOCK_NODE} {{
     {NEO4J_TASK_LOCK_TASK_ID}: $taskId 
    }})
WHERE lock.{NEO4J_TASK_LOCK_WORKER_ID} = $workerId
WITH lock
MATCH (t:{NEO4J_TASK_NODE} {{ {TASK_ID}: lock.{NEO4J_TASK_LOCK_TASK_ID} }})
CALL apoc.create.setLabels(t, $labels) YIELD node AS task
DELETE lock
RETURN task, lock
"""
    res = await tx.run(query, taskId=task_id, workerId=worker_id)
    try:
        await res.single(strict=True)
    except ResultNotSingleError as e:
        raise UnknownTask(task_id, worker_id) from e



import functools
import logging
from contextlib import asynccontextmanager
from copy import deepcopy
from inspect import iscoroutinefunction, signature
from typing import Callable, Dict, List, Optional, Tuple, Type, Union, final

from pydantic import validator
from typing_extensions import Self

from icij_common.pydantic_utils import ICIJModel, ICIJSettings
from icij_worker.routing_strategy import RoutingStrategy
from icij_worker.typing_ import Dependency
from icij_worker.utils import run_deps
from icij_worker.utils.imports import import_variable

logger = logging.getLogger(__name__)

PROGRESS_HANDLER_ARG = "progress"


class TaskGroup(ICIJModel):
    name: str
    timeout_s: Optional[int] = None
    max_task_queue_size: Optional[int] = None


class AsyncAppConfig(ICIJSettings):
    late_ack: bool = True
    recover_from_worker_timeout: bool = False
    max_task_queue_size: Optional[int] = None

    class Config:
        env_prefix = "ICIJ_APP_"


class RegisteredTask(ICIJModel):
    task: Callable
    recover_from: Tuple[Type[Exception], ...] = tuple()
    max_retries: Optional[int]
    group: Optional[TaskGroup]
    timeout_s: Optional[int]

    @validator("group", pre=True)
    def validate_group_instance(cls, v):  # pylint: disable=no-self-argument
        if isinstance(v, str):
            v = TaskGroup(name=v)
        return v


class AsyncApp:
    def __init__(
        self,
        name: str,
        config: AsyncAppConfig = None,
        dependencies: Optional[List[Dependency]] = None,
        routing_strategy: Optional[RoutingStrategy] = None,
    ):
        self._name = name
        if config is None:
            config = AsyncAppConfig()  # This will load from the env
        self._config = config
        self._registry = dict()
        self._groups = dict()
        if dependencies is None:
            dependencies = []
        self._dependencies = dependencies
        if routing_strategy is None:
            routing_strategy = RoutingStrategy()
        self._routing_strategy = routing_strategy

    @property
    def config(self) -> AsyncAppConfig:
        return self._config

    def with_config(self, value: AsyncAppConfig) -> Self:
        if not isinstance(value, AsyncAppConfig):
            raise TypeError(f"Expected {AsyncAppConfig.__name__}, got {value}")
        self._config = value
        return self

    @property
    def registry(self) -> Dict[str, RegisteredTask]:
        return self._registry

    @property
    def registered_keys(self) -> List[str]:
        return sorted(self._registry)

    @functools.cached_property
    def name(self) -> str:
        return self._name

    @functools.cached_property
    def routing_strategy(self) -> RoutingStrategy:
        return self._routing_strategy

    def with_routing_strategy(self, ns: RoutingStrategy) -> Self:
        self._routing_strategy = ns
        return self

    def task(
        self,
        name: Optional[str] = None,
        recover_from: Tuple[Type[Exception]] = tuple(),
        max_retries: Optional[int] = None,
        *,
        group: Optional[Union[str, TaskGroup]] = None,
    ) -> Callable:
        if callable(name) and not recover_from and max_retries is None:
            f = name
            return functools.partial(self._register_task, name=f.__name__, group=group)(
                f
            )
        if max_retries is None:
            max_retries = 3
        return functools.partial(
            self._register_task,
            name=name,
            recover_from=recover_from,
            max_retries=max_retries,
            group=group,
        )

    def task_group(self, name: str) -> Optional[TaskGroup]:
        return self._groups.get(name)

    @property
    def task_groups(self) -> List[TaskGroup]:
        return list(self._groups.values())

    @final
    @asynccontextmanager
    async def lifetime_dependencies(self, **kwargs):
        ctx = f"{self.name} async app"
        async with run_deps(self._dependencies, ctx=ctx, **kwargs):
            yield

    def _register_task(
        self,
        f: Callable,
        *,
        name: Optional[str] = None,
        recover_from: Tuple[Type[Exception]] = tuple(),
        max_retries: Optional[int] = None,
        group: Optional[Union[str, TaskGroup]] = None,
    ) -> Callable:
        if not iscoroutinefunction(f) and supports_progress(f):
            msg = (
                f"{f} is not a coroutine, progress is not supported as progress"
                f" reporting is inherently async, turn your function task into a"
                f" coroutine if necessary and use `await progress(my_progress)`"
            )
            raise ValueError(msg)
        if name is None:
            name = f.__name__
        registered = self._registry.get(name)
        if registered is not None:
            raise ValueError(f'Task "{name}" is already registered: {registered}')
        registered = RegisteredTask(
            task=f, max_retries=max_retries, recover_from=recover_from, group=group
        )
        self._validate_group(registered)
        self._registry[name] = registered
        if registered.group is not None:
            self._groups[registered.group.name] = registered.group

        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        return wrapped

    def _validate_group(self, task: RegisteredTask):
        if task.group is None:
            return
        existing = self._groups.get(task.group.name)
        if existing is not None and existing.name != task.group.name:
            msg = (
                f"invalid task group {task.group}, it has the same name as registered "
                f"group {existing}, use {existing} directly or specify a different name"
            )
            raise ValueError(msg)

    @classmethod
    def load(cls, app_path: str, config: Optional[AsyncAppConfig] = None) -> Self:
        app = deepcopy(import_variable(app_path))
        if config is not None:
            app.with_config(config)
        return app

    def filter_tasks(self, group: Optional[str]) -> Self:
        if group is None:
            return self
        kept = {
            t_name
            for t_name, t in self._registry.items()
            if self._routing_strategy.app_tasks_filter(
                task_group=t.group, app_group_name=group
            )
        }
        discarded = set(self._registry) - kept
        logger.info(
            "Applied group filtering:\n- running: %s\n- discarded: %s",
            ", ".join(sorted(kept)),
            ", ".join(sorted(discarded)),
        )
        self._registry = {k: self._registry[k] for k in kept}
        return self

    def __deepcopy__(self, memodict={}) -> Self:
        # pylint: disable=dangerous-default-value
        app = AsyncApp(
            name=self.name,
            config=deepcopy(self.config),
            dependencies=list(self._dependencies),
            routing_strategy=self.routing_strategy,
        )
        app._registry = deepcopy(self._registry)
        app._groups = deepcopy(self._groups)
        return app


def supports_progress(task_fn) -> bool:
    return any(
        param.name == PROGRESS_HANDLER_ARG
        for param in signature(task_fn).parameters.values()
    )



# pylint: disable=redefined-outer-name
# Test utils meant to be imported from clients libs to test their implem of workers
from __future__ import annotations

import asyncio
import logging
import signal
from abc import ABC
from datetime import datetime, timezone
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Tuple,
    TypeVar,
    cast,
)

from pydantic import Field

import icij_worker
from icij_common.pydantic_utils import (
    ICIJModel,
    IgnoreExtraModel,
)
from icij_worker import (
    AsyncApp,
    AsyncBackend,
    FSKeyValueStorage,
    ManagerEvent,
    RoutingStrategy,
    Task,
    TaskState,
    Worker,
    WorkerConfig,
)
from icij_worker.app import AsyncAppConfig, TaskGroup
from icij_worker.event_publisher import EventPublisher
from icij_worker.exceptions import (
    MessageDeserializationError,
    TaskQueueIsFull,
    UnknownTask,
)
from icij_worker.objects import (
    CancelEvent,
    Message,
    WorkerEvent,
)
from icij_worker.task_manager import TaskManager, TaskManagerConfig
from icij_worker.typing_ import RateProgress
from icij_worker.utils.dependencies import DependencyInjectionError
from icij_worker.utils.logging_ import LogWithWorkerIDMixin

logger = logging.getLogger(__name__)

_has_pytest = False  # necessary because of the pytest decorators which requires pytest
# to be defined
try:
    import pytest

    _has_pytest = True
except ImportError:
    pass

if _has_pytest:

    # TODO: make this one a MockStorage
    class DBMixin(FSKeyValueStorage, ABC):
        _locks_db_name = "locks"
        _manager_events_db_name = "manager_events"
        _worker_events_db_name = "worker_events"

        _routing_strategy: RoutingStrategy

        def __init__(self, db_path: Path) -> None:
            super().__init__(db_path)
            self._task_meta: Dict[str, Tuple[str, str]] = dict()

        def _make_group_dbs(self) -> Dict:
            dbs = dict()
            for k in [
                self._tasks_db_name,
                self._results_db_name,
                self._errors_db_name,
                self._locks_db_name,
                self._manager_events_db_name,
                self._worker_events_db_name,
            ]:
                dbs[k] = self._make_db(self._db_path, name=k)
            return dbs

        @property
        def db_path(self) -> Path:
            return self._db_path

        @staticmethod
        def _order_events(events: List[Dict]) -> float:
            events = [
                datetime.fromisoformat(evt["createdAt"]).timestamp() for evt in events
            ]
            return -min(events)

        async def _consume_(
            self,
            db_name: str,
            sleep_interval: float,
            factory: Callable[[Any], R],
            group: Optional[str] = None,
            select: Optional[Callable[[R], bool]] = None,
            order: Optional[Callable[[R], Any]] = None,
        ) -> R:
            while "i'm waiting until I find something interesting":
                selected = dict(self._dbs[db_name].items())
                if db_name == self._tasks_db_name:
                    selected = {
                        k: v
                        for k, v in selected.items()
                        if k not in self._dbs[self._locks_db_name]
                    }
                if group is not None:
                    selected = [
                        (k, t) for k, t in selected.items() if t.get("group") == group
                    ]
                    for k, t in selected:
                        t.pop("group", None)
                else:
                    selected = selected.items()
                selected = [(k, factory(t)) for k, t in selected]
                if select is not None:
                    selected = [(k, t) for k, t in selected if select(t)]
                if selected:
                    if order is not None:
                        k, t = min(selected, key=lambda x: order(x[1]))
                    else:
                        k, t = selected[0]
                    return t
                await asyncio.sleep(sleep_interval)

    class MockAppConfig(ICIJModel, LogWithWorkerIDMixin):
        # Just provide logging stuff to be able to see nice logs while doing TDD
        log_level: str = "DEBUG"
        loggers: List[str] = [icij_worker.__name__]

    _MOCKED_CONFIG: Optional[MockAppConfig] = None

    async def mock_async_config_enter(**_):
        global _MOCKED_CONFIG
        _MOCKED_CONFIG = MockAppConfig()
        logger.info("Loading mocked configuration %s", _MOCKED_CONFIG.json(indent=2))

    def lifespan_config() -> MockAppConfig:
        if _MOCKED_CONFIG is None:
            raise DependencyInjectionError("config")
        return _MOCKED_CONFIG

    def loggers_enter(worker_id: str, **_):
        config = lifespan_config()
        config.setup_loggers(worker_id=worker_id)
        logger.info("worker loggers ready to log 💬")

    mocked_app_deps = [
        ("configuration loading", mock_async_config_enter, None),
        ("loggers setup", loggers_enter, None),
    ]

    APP = AsyncApp(name="test-app", dependencies=mocked_app_deps)

    async def _hello_world(
        greeted: str, progress: Optional[RateProgress] = None
    ) -> str:
        if progress is not None:
            await progress(0.1)
        greeting = f"Hello {greeted} !"
        if progress is not None:
            await progress(0.99)
        return greeting

    @APP.task
    async def hello_world(greeted: str, progress: Optional[RateProgress] = None) -> str:
        return await _hello_world(greeted, progress)

    @APP.task(group="hello")
    async def grouped_hello_world(
        greeted: str, progress: Optional[RateProgress] = None
    ) -> str:
        return await _hello_world(greeted, progress)

    @APP.task
    def hello_world_sync(greeted: str) -> str:
        greeting = f"Hello {greeted} !"
        return greeting

    async def _sleep_for(
        duration: float, s: float = 0.01, progress: Optional[RateProgress] = None
    ):
        start = datetime.now()
        elapsed = 0
        while elapsed < duration:
            elapsed = (datetime.now() - start).total_seconds()
            await asyncio.sleep(s)
            if progress is not None:
                p = min(elapsed / duration, 1.0)
                await progress(p)

    @APP.task(max_retries=1)
    async def sleep_for(
        duration: float, s: float = 0.01, progress: Optional[RateProgress] = None
    ):
        await _sleep_for(duration, s, progress)

    short_timeout = 1
    short_tasks_group = TaskGroup(name="short", timeout_s=short_timeout)

    @APP.task(max_retries=3, group=short_tasks_group)
    async def sleep_for_short(
        duration: float, s: float = 0.01, progress: Optional[RateProgress] = None
    ):
        await _sleep_for(duration, s, progress)

    @APP.task(max_retries=666)
    async def often_retriable() -> str:
        pass

    @APP.task
    def case_test_task(snake_case_arg: str):
        return snake_case_arg

    @pytest.fixture(scope="session")
    def test_async_app() -> AsyncApp:
        return AsyncApp.load(f"{__name__}.APP")

    @pytest.fixture(scope="session")
    def app_config() -> AsyncAppConfig:
        return AsyncAppConfig()

    @pytest.fixture(scope="session")
    def late_ack_app_config() -> AsyncAppConfig:
        return AsyncAppConfig(late_ack=True)

    @pytest.fixture(scope="session")
    def test_async_app_late(late_ack_app_config: AsyncAppConfig) -> AsyncApp:
        return AsyncApp.load(f"{__name__}.APP", config=late_ack_app_config)

    class MockManagerConfig(TaskManagerConfig):
        backend: ClassVar[AsyncBackend] = Field(const=True, default=AsyncBackend.mock)
        db_path: Path
        event_refresh_interval_s: float = 0.1

    @TaskManager.register(AsyncBackend.mock)
    class MockManager(TaskManager, DBMixin):
        def __init__(
            self, app: AsyncApp, db_path: Path, event_refresh_interval_s: float = 0.1
        ):
            super().__init__(app)
            super(TaskManager, self).__init__(db_path)
            self._event_refresh_interval_s = event_refresh_interval_s

        @property
        def app(self) -> AsyncApp:
            return self._app

        async def _ensure_queue_size(self):
            tasks = await self.get_tasks(group=None, state=TaskState.QUEUED)
            n_queued = len(tasks)
            if (
                self.max_task_queue_size is not None
                and n_queued > self.max_task_queue_size
            ):
                raise TaskQueueIsFull(self.max_task_queue_size)

        async def _enqueue(self, task: Task):
            # pylint: disable=unused-argument
            await self._ensure_queue_size()

        async def _requeue(self, task: Task):
            await self._ensure_queue_size()
            group = await self.get_task_group(task.id)
            await self.save_task_(task, group)

        async def _consume(self) -> ManagerEvent:
            events = await self._consume_(
                self._manager_events_db_name,
                self._event_refresh_interval_s,
                list,
                select=lambda events: bool(  # pylint: disable=unnecessary-lambda
                    events
                ),
                order=self._order_events,
            )
            events = sorted(
                events,
                key=lambda e: datetime.fromisoformat(e["createdAt"]).timestamp(),
            )
            event = events[0]
            task_id = event["taskId"]
            key = self._key(task_id=task_id, obj_cls=ManagerEvent)
            manager_events_db = self._dbs[self._manager_events_db_name]
            manager_events_db[key] = events[1:]
            manager_events_db.commit(blocking=True)
            return cast(ManagerEvent, Message.parse_obj(event))

        async def cancel(self, task_id: str, *, requeue: bool):
            cancel_event = CancelEvent(
                task_id=task_id, requeue=requeue, created_at=datetime.now(timezone.utc)
            )
            key = self._key(task_id=task_id, obj_cls=CancelEvent)
            worker_events = self._dbs[self._worker_events_db_name]
            events = worker_events.get(key, [])
            event_dict = cancel_event.dict(exclude_unset=True, by_alias=True)
            events.append(event_dict)
            worker_events[key] = events
            worker_events.commit(blocking=True)

        @classmethod
        def _from_config(cls, config: MockManagerConfig, **extras) -> MockManager:
            tm = cls(
                config.app,
                config.db_path,
                event_refresh_interval_s=config.event_refresh_interval_s,
            )
            return tm

    R = TypeVar("R")

    class MockEventPublisher(DBMixin, EventPublisher):
        _excluded_from_event_update = {"error"}

        def __init__(self, db_path: Path):
            super().__init__(db_path)
            self.published_events = []

        async def _publish_event(self, event: ManagerEvent):
            key = self._key(event.task_id, ManagerEvent)
            manager_events_db = self._dbs[self._manager_events_db_name]
            event_dict = event.dict(exclude_unset=True, by_alias=True)
            events = manager_events_db.get(key, [])
            events.append(event_dict)
            manager_events_db[key] = events
            manager_events_db.commit(blocking=True)
            self.published_events.append(event)

    @WorkerConfig.register()
    class MockWorkerConfig(WorkerConfig, IgnoreExtraModel):
        type: ClassVar[str] = Field(const=True, default=AsyncBackend.mock.value)

        db_path: Path
        log_level: str = "DEBUG"
        loggers: List[str] = [icij_worker.__name__]
        task_queue_poll_interval_s: float = 2.0

    @Worker.register(AsyncBackend.mock)
    class MockWorker(Worker, MockEventPublisher):
        def __init__(
            self,
            app: AsyncApp,
            worker_id: Optional[str] = None,
            *,
            group: Optional[str],
            db_path: Path,
            poll_interval_s: float,
            **kwargs,
        ):
            MockEventPublisher.__init__(self, db_path)
            Worker.__init__(self, app, worker_id, group=group, **kwargs)
            self._poll_interval_s = poll_interval_s
            self._worker_id = worker_id
            self._logger_ = logging.getLogger(__name__)
            self.terminated_cancelled_event_loop = False

        @property
        def app(self) -> AsyncApp:
            return self._app

        @property
        def watch_cancelled_task(self) -> Optional[asyncio.Task]:
            return self._watch_cancelled_task

        async def work_forever_async(self):
            await self._work_forever_async()

        @property
        def work_forever_task(self) -> Optional[asyncio.Task]:
            return self._work_forever_task

        @property
        def work_once_task(self) -> Optional[asyncio.Task]:
            return self._work_once_task

        @property
        def successful_exit(self) -> bool:
            return self._successful_exit

        @property
        def current(self) -> Optional[Task]:
            return self._current

        @current.setter
        def current(self, value: Optional[Task]):
            self._current = value

        async def signal_handler(self, signal_name: signal.Signals, *, graceful: bool):
            await self._signal_handler(signal_name, graceful=graceful)

        async def _aenter__(self):
            if not Path(self._db_path).exists():
                raise OSError(f"worker DB was not initialized ({self._db_path})")

        @classmethod
        def _from_config(cls, config: MockWorkerConfig, **extras) -> MockWorker:
            worker = cls(
                db_path=config.db_path,
                poll_interval_s=config.task_queue_poll_interval_s,
                **extras,
            )
            return worker

        async def _acknowledge(self, task: Task):
            key = self._key(task.id, Task)
            locks = self._dbs[self._locks_db_name]
            try:
                del locks[key]
            except KeyError as e:
                raise UnknownTask(task.id) from e
            locks.commit()

        async def _negatively_acknowledge(self, nacked: Task):
            key = self._key(nacked.id, Task)
            locks = self._dbs[self._locks_db_name]
            del locks[key]
            locks.commit()

        @staticmethod
        def _task_factory(obj: Dict) -> Task:
            obj.pop("group", None)
            try:
                task = Task.parse_obj(obj)
            except Exception as e:
                msg = f"invalid task object {obj}"
                raise MessageDeserializationError(msg) from e
            return task

        async def _consume(self) -> Task:
            task = await self._consume_(
                self._tasks_db_name,
                self._poll_interval_s,
                self._task_factory,
                group=self._group,
                select=lambda t: t.state is TaskState.QUEUED,
                order=lambda t: t.created_at,
            )
            locks = self._dbs[self._locks_db_name]
            key = self._key(task.id, Task)
            locks[key] = self._id
            locks.commit()
            return task

        async def _consume_worker_events(self) -> WorkerEvent:
            events = await self._consume_(
                self._worker_events_db_name,
                self._poll_interval_s,
                list,
                group=self._group,
                select=lambda events: bool(  # pylint: disable=unnecessary-lambda
                    events
                ),
                order=self._order_events,
            )
            events = sorted(
                events,
                key=lambda e: datetime.fromisoformat(e["createdAt"]).timestamp(),
            )
            event = events[0]
            try:
                event = cast(WorkerEvent, Message.parse_obj(event))
            except Exception as e:
                msg = f"invalid event object {event}"
                raise MessageDeserializationError(msg) from e
            worker_events = self._dbs[self._worker_events_db_name]
            key = self._key(event.task_id, WorkerEvent)
            worker_events[key] = events[1:]
            worker_events.commit(blocking=True)
            return event

        async def work_once(self):
            await self._work_once()

        async def publish_cancelled_event(self, cancel_event: CancelEvent):
            await self._publish_cancelled_event(cancel_event)



from __future__ import annotations

from functools import lru_cache
from typing import Callable, Dict, Optional

from icij_common.pydantic_utils import LowerCamelCaseModel, NoEnumModel

try:
    from aio_pika import ExchangeType

    class Exchange(NoEnumModel, LowerCamelCaseModel):
        name: str
        type: ExchangeType

    class Routing(LowerCamelCaseModel):
        exchange: Exchange
        routing_key: str
        queue_name: str
        queue_args: Optional[Dict] = None
        dead_letter_routing: Optional[Routing] = None

except ImportError:
    pass

POSTGRES_DEFAULT = "postgres"


class RoutingStrategy:
    """Override this class to implement your own routing strategy between task groups
    and DBs, amqp queues and so on..."""

    def app_tasks_filter(
        self, *, task_group: Optional["TaskGroup"], app_group_name: str
    ) -> bool:
        """Used to filter app tasks so that the app can be started with a restricted
        group. Useful when tasks from the same app must be run by different workers
        """
        if task_group is None:
            return False
        return task_group.name.startswith(app_group_name)

    @classmethod
    @lru_cache
    def amqp_task_routing(cls, task_group: Optional[str]) -> Routing:
        """Used to route task with the right AMQP routing key based on the group"""
        # Overriding this default might require overriding AMQPTaskManager/AMQPWorker
        # so that they communicate correctly
        from icij_worker.utils.amqp import AMQPMixin

        default_task_routing = AMQPMixin.default_task_routing()
        exchange_name = default_task_routing.exchange.name
        routing_key = default_task_routing.routing_key
        queue_name = default_task_routing.queue_name
        if task_group is not None:
            routing_key = task_group
            queue_name += f".{task_group}"

        return Routing(
            exchange=Exchange(name=exchange_name, type=ExchangeType.DIRECT),
            routing_key=routing_key,
            queue_name=queue_name,
            queue_args=default_task_routing.queue_args,
            # TODO: route DLQ by group ???
            dead_letter_routing=default_task_routing.dead_letter_routing,
        )

    @classmethod
    def db_filter_factory(cls, worker_group: str) -> Callable[[str], bool]:
        """Used during DB task polling to poll only from the DBs supported by the
        worker group.

        This factory should return a callable which will take the DB name and
         return whether the DB is supported for that worker group
        """
        # pylint: disable=unused-argument
        # By default, workers are allowed to listen to all DBs
        return lambda db_name: True

    @classmethod
    def neo4j_db(cls, group: Optional[str]) -> str:
        # pylint: disable=unused-argument
        # By default, task from all groups are saved in the default neo4j db
        from icij_common.neo4j.db import NEO4J_COMMUNITY_DB

        return NEO4J_COMMUNITY_DB

    @classmethod
    def postgres_db(cls, group: Optional[str]) -> str:
        # pylint: disable=unused-argument
        return POSTGRES_DEFAULT

    @classmethod
    def test_db(cls, group: Optional[str]) -> str:
        # pylint: disable=unused-argument
        from icij_common.test_utils import TEST_DB

        return TEST_DB



from __future__ import annotations

from abc import ABC
from pathlib import Path
from typing import ClassVar, Optional

from pydantic import Field

from icij_worker.utils.registrable import RegistrableSettings


class WorkerConfig(RegistrableSettings, ABC):
    registry_key: ClassVar[str] = Field(const=True, default="type")

    # TODO: is app_dependencies_path better ?
    app_bootstrap_config_path: Optional[Path] = None
    inactive_after_s: Optional[float] = None
    log_level: str = "INFO"
    task_queue_poll_interval_s: float = 1.0
    type: ClassVar[str]

    class Config:
        env_prefix = "ICIJ_WORKER_"



import itertools
import json
from contextlib import asynccontextmanager
from copy import deepcopy
from datetime import datetime
from typing import AsyncGenerator, Dict, List, Optional, Tuple, Union

import neo4j
from neo4j import AsyncTransaction
from neo4j.exceptions import ResultNotSingleError

from icij_common.neo4j.db import db_specific_session
from icij_common.neo4j.migrate import retrieve_dbs
from icij_common.pydantic_utils import jsonable_encoder
from icij_worker.constants import (
    NEO4J_TASK_ARGS,
    NEO4J_TASK_ARGUMENTS_DEPRECATED,
    NEO4J_TASK_CANCEL_EVENT_CANCELLED_AT,
    NEO4J_TASK_CANCEL_EVENT_CREATED_AT_DEPRECATED,
    NEO4J_TASK_CANCEL_EVENT_NODE,
    NEO4J_TASK_COMPLETED_AT,
    NEO4J_TASK_CREATED_AT,
    NEO4J_TASK_ERROR_DETAIL_DEPRECATED,
    NEO4J_TASK_ERROR_ID_DEPRECATED,
    NEO4J_TASK_ERROR_MESSAGE,
    NEO4J_TASK_ERROR_NAME,
    NEO4J_TASK_ERROR_NODE,
    NEO4J_TASK_ERROR_OCCURRED_AT_DEPRECATED,
    NEO4J_TASK_ERROR_OCCURRED_TYPE,
    NEO4J_TASK_ERROR_OCCURRED_TYPE_OCCURRED_AT,
    NEO4J_TASK_ERROR_OCCURRED_TYPE_RETRIES_LEFT,
    NEO4J_TASK_ERROR_STACKTRACE,
    NEO4J_TASK_ERROR_TITLE_DEPRECATED,
    NEO4J_TASK_GROUP,
    NEO4J_TASK_HAS_RESULT_TYPE,
    NEO4J_TASK_ID,
    NEO4J_TASK_INPUTS_DEPRECATED,
    NEO4J_TASK_LOCK_NODE,
    NEO4J_TASK_LOCK_TASK_ID,
    NEO4J_TASK_LOCK_WORKER_ID,
    NEO4J_TASK_MANAGER_EVENT_NODE,
    NEO4J_TASK_MANAGER_EVENT_NODE_CREATED_AT,
    NEO4J_TASK_MAX_RETRIES,
    NEO4J_TASK_NAME,
    NEO4J_TASK_NAMESPACE_DEPRECATED,
    NEO4J_TASK_NODE,
    NEO4J_TASK_PROGRESS,
    NEO4J_TASK_RESULT_NODE,
    NEO4J_TASK_RESULT_RESULT,
    NEO4J_TASK_RETRIES_DEPRECATED,
    NEO4J_TASK_RETRIES_LEFT,
    NEO4J_TASK_TYPE_DEPRECATED,
)
from icij_worker.exceptions import MissingTaskResult, UnknownTask
from icij_worker.objects import ErrorEvent, ResultEvent, Task, TaskState, TaskUpdate
from icij_worker.task_storage import TaskStorage


class Neo4jStorage(TaskStorage):
    def __init__(self, driver: neo4j.AsyncDriver):
        self._driver = driver
        self._task_meta: Dict[str, Tuple[str, str]] = dict()

    async def get_task(self, task_id: str) -> Task:
        async with self._task_session(task_id) as sess:
            return await sess.execute_read(_get_task_tx, task_id=task_id)

    async def get_task_errors(self, task_id: str) -> List[ErrorEvent]:
        async with self._task_session(task_id) as sess:
            recs = await sess.execute_read(_get_task_errors_tx, task_id=task_id)
        errors = [ErrorEvent.from_neo4j(rec) for rec in recs]
        return errors

    async def get_task_result(self, task_id: str) -> ResultEvent:
        async with self._task_session(task_id) as sess:
            return await sess.execute_read(_get_task_result_tx, task_id=task_id)

    async def get_tasks(
        self,
        group: Optional[str],
        *,
        task_name: Optional[str] = None,
        state: Optional[Union[List[TaskState], TaskState]] = None,
        **kwargs,
    ) -> List[Task]:
        db = self._routing_strategy.neo4j_db(group)
        async with self._db_session(db) as sess:
            recs = await _get_tasks(sess, state=state, task_name=task_name, group=group)
        tasks = [Task.from_neo4j(r) for r in recs]
        return tasks

    async def get_task_group(self, task_id: str) -> Optional[str]:
        if task_id not in self._task_meta:
            await self._refresh_task_meta()
        try:
            return self._task_meta[task_id][1]
        except KeyError as e:
            raise UnknownTask(task_id) from e

    async def save_task_(self, task: Task, group: Optional[str]) -> bool:
        db = self._routing_strategy.neo4j_db(group)
        async with self._db_session(db) as sess:
            task_props = task.dict(by_alias=True, exclude_unset=True)
            new_task = await sess.execute_write(
                _save_task_tx,
                task_id=task.id,
                task_props=task_props,
                group=group,
            )

        self._task_meta[task.id] = (db, group)
        return new_task

    async def save_result(self, result: ResultEvent):
        async with self._task_session(result.task_id) as sess:
            res_str = json.dumps(jsonable_encoder(result.result))
            await sess.execute_write(
                _save_result_tx,
                task_id=result.task_id,
                result=res_str,
                completed_at=result.created_at,
            )

    async def save_error(self, error: ErrorEvent):
        async with self._task_session(error.task_id) as sess:
            error_props = error.error.dict(by_alias=True)
            error_props.pop("@type")
            error_props["stacktrace"] = [
                json.dumps(item) for item in error_props["stacktrace"]
            ]
            await sess.execute_write(
                _save_error_tx,
                task_id=error.task_id,
                error_props=error_props,
                retries_left=error.retries_left,
                created_at=error.created_at,
            )

    @asynccontextmanager
    async def _db_session(self, db: str) -> AsyncGenerator[neo4j.AsyncSession, None]:
        async with db_specific_session(self._driver, db) as sess:
            yield sess

    @asynccontextmanager
    async def _task_session(
        self, task_id: str
    ) -> AsyncGenerator[neo4j.AsyncSession, None]:
        db = await self._get_task_db(task_id)
        async with self._db_session(db) as sess:
            yield sess

    async def _get_task_db(self, task_id: str) -> str:
        if task_id not in self._task_meta:
            await self._refresh_task_meta()
        try:
            return self._task_meta[task_id][0]
        except KeyError as e:
            raise UnknownTask(task_id) from e

    async def _refresh_task_meta(self):
        dbs = await retrieve_dbs(self._driver)
        for db in dbs:
            async with self._db_session(db.name) as sess:
                # Here we make the assumption that task IDs are unique across
                # projects and not per project
                task_meta = {
                    meta["taskId"]: (db.name, meta["taskNs"])
                    for meta in await sess.execute_read(_get_tasks_meta_tx)
                }
                self._task_meta.update(task_meta)


async def _get_tasks_meta_tx(tx: neo4j.AsyncTransaction) -> List[neo4j.Record]:
    query = f"""MATCH (task:{NEO4J_TASK_NODE})
RETURN task.{NEO4J_TASK_ID} as taskId, task.{NEO4J_TASK_GROUP} as taskNs"""
    res = await tx.run(query)
    meta = [rec async for rec in res]
    return meta


async def _save_task_tx(
    tx: neo4j.AsyncTransaction,
    *,
    task_id: str,
    task_props: Dict,
    group: Optional[str],
) -> bool:
    query = f"MATCH (task:{NEO4J_TASK_NODE} {{{NEO4J_TASK_ID}: $taskId }}) RETURN task"
    res = await tx.run(query, taskId=task_id)
    existing = None
    task_props = deepcopy(task_props)
    try:
        existing = await res.single(strict=True)
    except ResultNotSingleError:
        task_props[NEO4J_TASK_ARGS] = json.dumps(
            task_props.get(NEO4J_TASK_ARGS, dict())
        )
        task_props[NEO4J_TASK_GROUP] = group
    else:
        task_obj = {"id": task_id}
        task_obj.update(task_props)
        task_props = TaskUpdate.from_task(Task.parse_obj(task_obj)).dict(
            by_alias=True, exclude_unset=True
        )
    task_props.pop("@type", None)
    if existing is not None and existing["task"]["group"] != group:
        msg = (
            f"DB task group ({existing['task']['group']}) differs from"
            f" save task group: {group}"
        )
        raise ValueError(msg)
    query = f"""MERGE (t:{NEO4J_TASK_NODE} {{{NEO4J_TASK_ID}: $taskId }})
SET t += $taskProps
WITH t
CALL apoc.create.setLabels(t, $labels) YIELD node AS task
RETURN task"""
    state = task_props.pop("state")
    labels = [NEO4J_TASK_NODE, state.value]
    await tx.run(query, taskId=task_id, taskProps=task_props, labels=labels)
    return existing is None


async def _save_result_tx(
    tx: neo4j.AsyncTransaction, *, task_id: str, result: str, completed_at: datetime
):
    query = f"""MATCH (t:{NEO4J_TASK_NODE} {{{NEO4J_TASK_ID}: $taskId }})
SET t.{NEO4J_TASK_PROGRESS} = 1.0, t.{NEO4J_TASK_COMPLETED_AT} = $completedAt
WITH t
CALL apoc.create.setLabels(t, $labels) YIELD node AS task
MERGE (task)-[:{NEO4J_TASK_HAS_RESULT_TYPE}]->(result:{NEO4J_TASK_RESULT_NODE})
ON CREATE SET result.{NEO4J_TASK_RESULT_RESULT} = $result
RETURN task, result"""
    labels = [NEO4J_TASK_NODE, TaskState.DONE.value]
    res = await tx.run(
        query, taskId=task_id, result=result, labels=labels, completedAt=completed_at
    )
    records = [rec async for rec in res]
    summary = await res.consume()
    if not records:
        raise UnknownTask(task_id)
    if not summary.counters.relationships_created:
        msg = f"Attempted to save result for task {task_id} but found existing result"
        raise ValueError(msg)


async def _save_error_tx(
    tx: neo4j.AsyncTransaction,
    task_id: str,
    *,
    error_props: Dict,
    retries_left: int,
    created_at: datetime,
):
    query = f"""MATCH (task:{NEO4J_TASK_NODE} {{{NEO4J_TASK_ID}: $taskId }})
CREATE (error:{NEO4J_TASK_ERROR_NODE})-[rel:{NEO4J_TASK_ERROR_OCCURRED_TYPE}]->(task)
SET error += $errorProps,
    rel.{NEO4J_TASK_ERROR_OCCURRED_TYPE_OCCURRED_AT} = $occurredAt,
    rel.{NEO4J_TASK_ERROR_OCCURRED_TYPE_RETRIES_LEFT} = $retriesLeft  
RETURN task, error"""
    res = await tx.run(
        query,
        taskId=task_id,
        errorProps=error_props,
        retriesLeft=retries_left,
        occurredAt=created_at,
    )
    try:
        await res.single(strict=True)
    except ResultNotSingleError as e:
        raise UnknownTask(task_id) from e


async def add_support_for_async_task_tx(tx: neo4j.AsyncTransaction):
    constraint_query = f"""CREATE CONSTRAINT constraint_task_unique_id
IF NOT EXISTS 
FOR (task:{NEO4J_TASK_NODE})
REQUIRE (task.{NEO4J_TASK_ID}) IS UNIQUE"""
    await tx.run(constraint_query)
    created_at_query = f"""CREATE INDEX index_task_created_at IF NOT EXISTS
FOR (task:{NEO4J_TASK_NODE})
ON (task.{NEO4J_TASK_CREATED_AT})"""
    await tx.run(created_at_query)
    type_query = f"""CREATE INDEX index_task_name IF NOT EXISTS
FOR (task:{NEO4J_TASK_NODE})
ON (task.{NEO4J_TASK_NAME})"""
    await tx.run(type_query)
    error_timestamp_query = f"""CREATE INDEX index_task_error_timestamp IF NOT EXISTS
FOR (task:{NEO4J_TASK_ERROR_NODE})
ON (task.{NEO4J_TASK_ERROR_OCCURRED_AT_DEPRECATED})"""
    await tx.run(error_timestamp_query)
    error_id_query = f"""CREATE CONSTRAINT constraint_task_error_unique_id IF NOT EXISTS
FOR (task:{NEO4J_TASK_ERROR_NODE})
REQUIRE (task.{NEO4J_TASK_ERROR_ID_DEPRECATED}) IS UNIQUE"""
    await tx.run(error_id_query)
    task_lock_task_id_query = f"""CREATE CONSTRAINT constraint_task_lock_unique_task_id
IF NOT EXISTS
FOR (lock:{NEO4J_TASK_LOCK_NODE})
REQUIRE (lock.{NEO4J_TASK_LOCK_TASK_ID}) IS UNIQUE"""
    await tx.run(task_lock_task_id_query)
    task_lock_worker_id_query = f"""CREATE INDEX index_task_lock_worker_id IF NOT EXISTS
FOR (lock:{NEO4J_TASK_LOCK_NODE})
ON (lock.{NEO4J_TASK_LOCK_WORKER_ID})"""
    await tx.run(task_lock_worker_id_query)


async def _get_tasks(
    sess: neo4j.AsyncSession,
    state: Optional[Union[List[TaskState], TaskState]],
    task_name: Optional[str],
    group: Optional[str],
) -> List[neo4j.Record]:
    if isinstance(state, TaskState):
        state = [state]
    if state is not None:
        state = [s.value for s in state]
    return await sess.execute_read(
        _get_tasks_tx, state=state, task_name=task_name, group=group
    )


async def _get_task_tx(tx: neo4j.AsyncTransaction, *, task_id: str) -> Task:
    query = f"MATCH (task:{NEO4J_TASK_NODE} {{ {NEO4J_TASK_ID}: $taskId }}) RETURN task"
    res = await tx.run(query, taskId=task_id)
    tasks = [Task.from_neo4j(t) async for t in res]
    if not tasks:
        raise UnknownTask(task_id)
    return tasks[0]


async def _get_tasks_tx(
    tx: neo4j.AsyncTransaction,
    state: Optional[List[str]],
    *,
    task_name: Optional[str],
    group: Optional[str],
) -> List[neo4j.Record]:
    where = ""
    if task_name:
        where = f"WHERE task.{NEO4J_TASK_NAME} = $type"
    if group is not None:
        if not where:
            where = "WHERE "
        else:
            where += " AND "
        where += f"task.{NEO4J_TASK_GROUP} = $group"
    all_labels = [(NEO4J_TASK_NODE,)]
    if isinstance(state, str):
        state = (state,)
    if state is not None:
        all_labels.append(tuple(state))
    all_labels = list(itertools.product(*all_labels))
    if all_labels:
        query = "UNION\n".join(
            f"""MATCH (task:{':'.join(labels)}) {where}
            RETURN task
            ORDER BY task.{NEO4J_TASK_CREATED_AT} DESC"""
            for labels in all_labels
        )
    else:
        query = f"""MATCH (task:{NEO4J_TASK_NODE})
RETURN task
ORDER BY task.{NEO4J_TASK_CREATED_AT} DESC"""
    res = await tx.run(query, type=task_name, group=group)
    recs = [rec async for rec in res]
    return recs


async def _get_task_errors_tx(
    tx: neo4j.AsyncTransaction, *, task_id: str
) -> List[neo4j.Record]:
    query = f"""MATCH (task:{NEO4J_TASK_NODE} {{ {NEO4J_TASK_ID}: $taskId }})
MATCH (error:{NEO4J_TASK_ERROR_NODE})-[rel:{NEO4J_TASK_ERROR_OCCURRED_TYPE}]->(task)
RETURN error, rel, task
ORDER BY rel.{NEO4J_TASK_ERROR_OCCURRED_TYPE_OCCURRED_AT} DESC
"""
    res = await tx.run(query, taskId=task_id)
    errors = [err async for err in res]
    return errors


async def _get_task_result_tx(
    tx: neo4j.AsyncTransaction, *, task_id: str
) -> ResultEvent:
    query = f"""MATCH (task:{NEO4J_TASK_NODE} {{ {NEO4J_TASK_ID}: $taskId }})
MATCH (task)-[:{NEO4J_TASK_HAS_RESULT_TYPE}]->(result:{NEO4J_TASK_RESULT_NODE})
RETURN task, result
"""
    res = await tx.run(query, taskId=task_id)
    results = [ResultEvent.from_neo4j(t) async for t in res]
    if not results:
        raise MissingTaskResult(task_id)
    return results[0]


async def migrate_task_errors_v0_tx(tx: neo4j.AsyncTransaction):
    query = f"""MATCH (error:{NEO4J_TASK_ERROR_NODE})
// We leave the stacktrace and cause empty
SET error.{NEO4J_TASK_ERROR_NAME} = error.{NEO4J_TASK_ERROR_TITLE_DEPRECATED},
    error.{NEO4J_TASK_ERROR_MESSAGE} = error.{NEO4J_TASK_ERROR_DETAIL_DEPRECATED},
    error.{NEO4J_TASK_ERROR_STACKTRACE} = []
REMOVE error.{NEO4J_TASK_ERROR_TITLE_DEPRECATED}, error.{NEO4J_TASK_ERROR_DETAIL_DEPRECATED}
RETURN error
"""
    await tx.run(query)


async def migrate_cancelled_event_created_at_v0_tx(tx: neo4j.AsyncTransaction):
    query = f"""MATCH (event:{NEO4J_TASK_CANCEL_EVENT_NODE})
SET event.{NEO4J_TASK_CANCEL_EVENT_CANCELLED_AT} 
    = event.{NEO4J_TASK_CANCEL_EVENT_CREATED_AT_DEPRECATED}
REMOVE event.{NEO4J_TASK_CANCEL_EVENT_CREATED_AT_DEPRECATED}
RETURN event
"""
    await tx.run(query)


async def migrate_add_index_to_task_namespace_v0_tx(tx: neo4j.AsyncTransaction):
    create_index = f"""
CREATE INDEX index_task_namespace IF NOT EXISTS
FOR (task:{NEO4J_TASK_NAMESPACE_DEPRECATED})
ON (task.{NEO4J_TASK_NAMESPACE_DEPRECATED})
"""
    await tx.run(create_index)


# pylint: disable=line-too-long
async def migrate_task_inputs_to_arguments_v0_tx(tx: neo4j.AsyncTransaction):
    query = f"""MATCH (task:{NEO4J_TASK_NODE})
SET task.{NEO4J_TASK_ARGS} = task.{NEO4J_TASK_INPUTS_DEPRECATED}
REMOVE task.{NEO4J_TASK_INPUTS_DEPRECATED}
RETURN task
"""
    await tx.run(query)


async def _rename_task_type_into_name_tx(tx: neo4j.AsyncTransaction):
    query = f"""MATCH (task:{NEO4J_TASK_NODE})
SET task.{NEO4J_TASK_NAME} = task.{NEO4J_TASK_TYPE_DEPRECATED}
REMOVE task.{NEO4J_TASK_TYPE_DEPRECATED}
RETURN task
"""
    await tx.run(query)


async def migrate_task_type_to_name_v0(sess: neo4j.AsyncSession):
    drop_index = "DROP INDEX index_task_type IF EXISTS"
    await sess.run(drop_index)
    create_index = f"""CREATE INDEX index_task_name IF NOT EXISTS
FOR (task:{NEO4J_TASK_NODE})
ON (task.{NEO4J_TASK_NAME})
"""
    await sess.run(create_index)
    await sess.execute_write(_rename_task_type_into_name_tx)


async def migrate_task_progress_v0_tx(tx: neo4j.AsyncTransaction):
    query = f"""MATCH (task:{NEO4J_TASK_NODE})
SET task.{NEO4J_TASK_PROGRESS} = toFloat(task.{NEO4J_TASK_PROGRESS}) / 100.0
RETURN task
"""
    await tx.run(query)


async def migrate_index_event_dates_v0_tx(tx: neo4j.AsyncTransaction):
    manager_event_query = f"""CREATE INDEX index_manager_events_created_at IF NOT EXISTS
FOR (event:{NEO4J_TASK_MANAGER_EVENT_NODE})
ON (event.{NEO4J_TASK_MANAGER_EVENT_NODE_CREATED_AT})"""
    await tx.run(manager_event_query)
    worker_event_query = f"""CREATE INDEX index_canceled_events_created_at IF NOT EXISTS
FOR (event:{NEO4J_TASK_CANCEL_EVENT_NODE})
ON (event.{NEO4J_TASK_CANCEL_EVENT_CANCELLED_AT})"""
    await tx.run(worker_event_query)


async def migrate_task_retries_and_error_v0_tx(
    tx: neo4j.AsyncTransaction,
):
    # Sadly, without the max retries save in DB, we can't compute the retries left, so
    # we just delete this attribute
    query = f"""MATCH (task:{NEO4J_TASK_NODE})
WHERE task.{NEO4J_TASK_RETRIES_LEFT} IS NULL
SET task.{NEO4J_TASK_MAX_RETRIES} = 3,
    task.{NEO4J_TASK_RETRIES_LEFT} = 3 - coalesce(task.{NEO4J_TASK_RETRIES_DEPRECATED}, 0)  
REMOVE task.{NEO4J_TASK_RETRIES_DEPRECATED}
RETURN task
"""
    await tx.run(query)
    query = f"""MATCH (error:{NEO4J_TASK_ERROR_NODE})-[rel:{NEO4J_TASK_ERROR_OCCURRED_TYPE}]-(task)
WHERE rel.{NEO4J_TASK_ERROR_OCCURRED_TYPE_OCCURRED_AT} IS NULL
SET rel.{NEO4J_TASK_ERROR_OCCURRED_TYPE_OCCURRED_AT} 
    = error.{NEO4J_TASK_ERROR_OCCURRED_AT_DEPRECATED},
    rel.{NEO4J_TASK_ERROR_OCCURRED_TYPE_RETRIES_LEFT} = 3
REMOVE error.{NEO4J_TASK_ERROR_OCCURRED_AT_DEPRECATED}, error.{NEO4J_TASK_ERROR_ID_DEPRECATED}
RETURN error
"""
    await tx.run(query)


async def migrate_task_arguments_into_args_v0_tx(
    tx: neo4j.AsyncTransaction,
):
    query = f"""MATCH (task:{NEO4J_TASK_NODE})
WHERE task.{NEO4J_TASK_ARGUMENTS_DEPRECATED} IS NOT NULL
SET task.{NEO4J_TASK_ARGS} = task.{NEO4J_TASK_ARGUMENTS_DEPRECATED}  
REMOVE task.{NEO4J_TASK_ARGUMENTS_DEPRECATED}
RETURN task
"""
    await tx.run(query)


async def migrate_task_namespace_into_group_v0(sess: neo4j.AsyncSession):
    drop_index = "DROP INDEX index_task_namespace IF EXISTS"
    await sess.run(drop_index)
    create_index = f"""CREATE INDEX index_task_group IF NOT EXISTS
FOR (task:{NEO4J_TASK_NODE})
ON (task.{NEO4J_TASK_GROUP})
"""
    await sess.run(create_index)
    await sess.execute_write(_rename_namespace_into_group_v0_tx)


async def _rename_namespace_into_group_v0_tx(tx: AsyncTransaction):
    query = f"""MATCH (task:{NEO4J_TASK_NODE})
WHERE task.{NEO4J_TASK_NAMESPACE_DEPRECATED} IS NOT NULL
SET task.{NEO4J_TASK_GROUP} = task.{NEO4J_TASK_NAMESPACE_DEPRECATED}  
REMOVE task.{NEO4J_TASK_NAMESPACE_DEPRECATED}
RETURN task
"""
    await tx.run(query)


# pylint: disable=line-too-long
MIGRATIONS = [
    add_support_for_async_task_tx,
    migrate_task_errors_v0_tx,
    migrate_cancelled_event_created_at_v0_tx,
    migrate_add_index_to_task_namespace_v0_tx,
    migrate_task_inputs_to_arguments_v0_tx,
    migrate_task_type_to_name_v0,
    migrate_task_progress_v0_tx,
    migrate_index_event_dates_v0_tx,
    migrate_task_retries_and_error_v0_tx,
    migrate_task_arguments_into_args_v0_tx,
    migrate_task_namespace_into_group_v0,
]



import asyncio
import functools
import logging
import signal
from abc import ABC
from asyncio import AbstractEventLoop
from typing import Optional

from icij_common.logging_utils import LogWithNameMixin

_HANDLE_SIGNALS = [signal.SIGTERM]


# TODO: rename this file to signals
class HandleSignalsMixin(LogWithNameMixin, ABC):
    _work_forever_task: Optional[asyncio.Task]
    _worker_cancelled: bool = False
    _loop: AbstractEventLoop
    cancel_lock: asyncio.Lock

    def __init__(self, logger: logging.Logger, handle_signals: bool = True):
        super().__init__(logger)
        self._handle_signals = handle_signals

    async def _aenter__(self):
        # TODO: define this one on the worker side
        if self._handle_signals:
            self._setup_child_process_signal_handlers()

    async def _signal_handler(self, signal_name: signal.Signals, *, graceful: bool):
        async with self.cancel_lock:
            self._worker_cancelled = True
            self.exception("received %s", signal_name)
            self._graceful_shutdown = graceful
            if self._work_forever_task is not None:
                self.info("cancelling worker loop...")
                self._work_forever_task.cancel()

    def _setup_child_process_signal_handlers(self):
        # We ignore SIGINT (graceful shutdown), this signal is handled by the
        # process handling the pool, which will terminate the pool and send a SIGTERM,
        # which is handled here

        self._loop.add_signal_handler(signal.SIGINT, signal.getsignal(signal.SIG_IGN))
        for s in _HANDLE_SIGNALS:
            handler = functools.partial(self._signal_handler, s, graceful=True)
            self._loop.add_signal_handler(s, handler)



import logging
import subprocess
from pathlib import Path
from typing import Optional

from .connection_info import PostgresConnectionInfo

MIGRATIONS_DIR = Path(__file__).parent.joinpath("db", "migrations")

logger = logging.getLogger(__name__)

_DEFAULT_MIGRATE_ARG = ["--no-dump-schema", "--wait", "-d", str(MIGRATIONS_DIR)]


def migrate(
    connection_info: PostgresConnectionInfo,
    db_name: str,
    timeout_s: Optional[float] = None,
):
    url = connection_info.url(db=db_name)

    global_args = list(_DEFAULT_MIGRATE_ARG)
    global_args.extend(("--url", url))
    if timeout_s is not None:
        global_args.extend(("--wait-timeout", f"{timeout_s:.0f}s"))
    args = ["dbmate"] + global_args + ["up", "--strict", "-v"]
    logger.debug("starting migration command: %s", " ".join(args))
    try:
        process = subprocess.run(
            args, capture_output=True, check=True, encoding="utf-8"
        )
    except subprocess.CalledProcessError as e:
        logger.error(
            'command "%s" failed with:\n- error: %s- output: %s',
            " ".join(e.cmd),
            e.stderr,
            e.output,
        )
        raise
    logger.debug("migration successful !\nLogs: \n%s", process.stdout)



import inspect
import logging
from contextlib import asynccontextmanager, contextmanager
from typing import AsyncGenerator, List

import sys

from icij_worker.typing_ import Dependency

logger = logging.getLogger(__name__)


class DependencyInjectionError(RuntimeError):
    def __init__(self, name: str):
        msg = f"{name} was not injected"
        super().__init__(msg)


@contextmanager
def _log_exception_and_continue():
    try:
        yield
    except Exception as exc:  # pylint: disable=broad-exception-caught
        logger.exception("Exception %s occurred ", exc)


@asynccontextmanager
async def run_deps(
    dependencies: List[Dependency], ctx: str, **kwargs
) -> AsyncGenerator[None, None]:
    to_close = []
    original_ex = None
    try:
        logger.info("Setting up dependencies for %s...", ctx)
        for name, enter_fn, exit_fn in dependencies:
            if enter_fn is not None:
                if name is not None:
                    logger.debug("Setup up dependency: %s", name)
                if inspect.iscoroutinefunction(enter_fn):
                    await enter_fn(**kwargs)
                else:
                    enter_fn(**kwargs)
            to_close.append((name, exit_fn))
        yield
    except Exception as e:  # pylint: disable=broad-exception-caught
        original_ex = e
    finally:
        to_raise = []
        if original_ex is not None:
            to_raise.append(original_ex)
        logger.info("Rolling back dependencies for %s...", ctx)
        for name, exit_fn in to_close[::-1]:
            if exit_fn is None:
                continue
            try:
                if name is not None:
                    logger.debug("rolling back %s", name)
                exc_info = sys.exc_info()
                with _log_exception_and_continue():
                    if inspect.iscoroutinefunction(exit_fn):
                        await exit_fn(*exc_info)
                    else:
                        exit_fn(*exc_info)
            except Exception as e:  # pylint: disable=broad-exception-caught
                to_raise.append(e)
        logger.debug("Rolled back all dependencies for %s !", ctx)
        if to_raise:
            for e in to_raise:
                logger.exception("Error while handling dependencies %s!", e)
            raise RuntimeError(to_raise) from to_raise[0]



import asyncio
import logging
from typing import List

logger = logging.getLogger(__name__)


class Seq:
    pass


def stop_other_tasks_when_exc(t: asyncio.Task, others: List[asyncio.Task]) -> None:
    try:
        exc = t.exception()
    except asyncio.CancelledError:
        msg = (
            "%s was cancelled, probably due to an exception in a concurrent infinite"
            " loops"
        )
        logger.debug(msg, t)
        return
    if exc is None:
        msg = "attempted to cancel %s, which was already done"
        logger.debug(msg, t)
        return
    loop = t.get_loop()
    others = [task for task in others if t is not task]
    for task in others:
        if not task.done():
            task.cancel(f"cancelled by {t}")
    loop.call_soon(_raise(exc))


def _raise(exc: Exception):
    logger.debug("exception occurred in one of the running infinite loops: %s", exc)
    raise exc



from __future__ import annotations

import asyncio
import re
from contextlib import AbstractAsyncContextManager, AsyncExitStack
from copy import deepcopy
from enum import Enum, unique
from functools import cached_property, lru_cache
from typing import Any, Dict, List, Optional, Tuple, Type, cast
from urllib import parse

import aiormq
from aio_pika import (
    DeliveryMode,
    Message as AioPikaMessage,
    RobustChannel as RobustChannel_,
    RobustConnection as RobustConnection_,
)
from aio_pika.abc import (
    AbstractExchange,
    AbstractQueueIterator,
    AbstractRobustChannel,
    AbstractRobustConnection,
    ExchangeType,
    TimeoutType,
    UnderlayConnection,
)
from aio_pika.exceptions import ChannelPreconditionFailed
from aiohttp import BasicAuth
from aiormq import Connection as AiormqConnection
from aiormq.abc import ArgumentsType, ConfirmationFrameType, URLorStr
from mdurl import URL
from pamqp.commands import Basic
from pamqp.common import FieldTable

from icij_common.pydantic_utils import ICIJModel, NoEnumModel
from icij_worker import Message
from icij_worker.app import TaskGroup
from icij_worker.constants import (
    AMQP_MANAGER_EVENTS_DL_QUEUE,
    AMQP_MANAGER_EVENTS_DL_ROUTING_KEY,
    AMQP_MANAGER_EVENTS_DL_X,
    AMQP_MANAGER_EVENTS_QUEUE,
    AMQP_MANAGER_EVENTS_ROUTING_KEY,
    AMQP_MANAGER_EVENTS_X,
    AMQP_TASKS_DL_QUEUE,
    AMQP_TASKS_DL_ROUTING_KEY,
    AMQP_TASKS_DL_X,
    AMQP_TASKS_QUEUE,
    AMQP_TASKS_ROUTING_KEY,
    AMQP_TASKS_X,
    AMQP_TASK_QUEUE_PRIORITY,
    AMQP_WORKER_EVENTS_QUEUE,
    AMQP_WORKER_EVENTS_ROUTING_KEY,
    AMQP_WORKER_EVENTS_X,
)
from icij_worker.exceptions import WorkerTimeoutError
from icij_worker.routing_strategy import Exchange, Routing, RoutingStrategy
from icij_worker.task_storage.postgres.postgres import logger
from icij_worker.utils.http import AiohttpClient

_DELIVERY_ACK_TIMEOUT_RE = re.compile(
    r"delivery acknowledgement on channel .+ timed out", re.MULTILINE
)


@unique
class ApplyTo(str, Enum):
    EXCHANGES = "exchanges"
    QUEUES = "queues"
    CLASSIC_QUEUES = "classic_queues"
    QUORUM_QUEUES = "quorum_queues"
    STREAMS = "streams"
    ALL = "all"


class AMQPPolicy(NoEnumModel):
    name: str
    pattern: str
    definition: Dict[str, Any]
    apply_to: Optional[ApplyTo] = None
    priority: Optional[int] = None


_DELIVERY_ACK_TIMEOUT_RE = re.compile(
    r"delivery acknowledgement on channel .+ timed out", re.MULTILINE
)


class AMQPConfigMixin(ICIJModel):
    connection_timeout_s: float = 5.0
    reconnection_wait_s: float = 5.0
    rabbitmq_host: str = "127.0.0.1"
    rabbitmq_password: str = "guest"
    rabbitmq_port: Optional[int] = 5672
    rabbitmq_management_port: Optional[int] = 15672
    rabbitmq_user: Optional[str] = "guest"
    rabbitmq_vhost: Optional[str] = "%2F"
    rabbitmq_is_qpid: bool = False

    @cached_property
    def broker_url(self) -> str:
        amqp_userinfo = self.rabbitmq_user
        if self.rabbitmq_password:
            amqp_userinfo += f":{self.rabbitmq_password}"
        if amqp_userinfo:
            amqp_userinfo += "@"
        amqp_authority = (
            f"{amqp_userinfo or ''}{self.rabbitmq_host}"
            f"{f':{self.rabbitmq_port}' or ''}"
        )
        amqp_uri = f"amqp://{amqp_authority}"
        if self.rabbitmq_vhost is not None:
            amqp_uri += f"/{self.rabbitmq_vhost}"
        return amqp_uri

    @cached_property
    def management_url(self) -> str:
        management_url = f"http://{self.rabbitmq_host}:{self.rabbitmq_management_port}"
        return management_url

    @cached_property
    def basic_auth(self) -> BasicAuth:
        return BasicAuth(self.rabbitmq_user, self.rabbitmq_password)

    def to_management_client(self) -> AMQPManagementClient:
        client = AMQPManagementClient(
            self.management_url,
            rabbitmq_vhost=self.rabbitmq_vhost,
            rabbitmq_auth=self.basic_auth,
        )
        return client


class AMQPMixin:
    _app_id: str
    _channel_: AbstractRobustChannel
    _routing_strategy: RoutingStrategy
    _task_x: AbstractExchange
    max_task_queue_size: Optional[int]
    _always_include = {"createdAt", "retriesLeft"}

    def __init__(
        self,
        broker_url: str,
        *,
        connection_timeout_s: float = 1.0,
        reconnection_wait_s: float = 5.0,
        inactive_after_s: float = None,
        is_qpid: bool = False,
    ):
        self._is_qpid = is_qpid
        self._broker_url = broker_url
        self._reconnection_wait_s = reconnection_wait_s
        self._connection_timeout_s = connection_timeout_s
        self._inactive_after_s = inactive_after_s
        self._publisher_confirms = not is_qpid
        self._connection_: Optional[AbstractRobustConnection] = None
        self._exit_stack = AsyncExitStack()

    async def _publish_message(
        self,
        message: Message,
        *,
        exchange: AbstractExchange,
        delivery_mode: DeliveryMode = DeliveryMode.PERSISTENT,
        routing_key: Optional[str],
        mandatory: bool,
    ) -> Optional[ConfirmationFrameType]:
        message = message.json(
            exclude_unset=True, by_alias=True, exclude_none=True
        ).encode()
        message = AioPikaMessage(
            message, delivery_mode=delivery_mode, app_id=self._app_id
        )
        confirmation = await exchange.publish(message, routing_key, mandatory=mandatory)
        if not isinstance(confirmation, Basic.Ack) and self._publisher_confirms:
            msg = f"Failed to deliver {message.body}, received {confirmation}"
            raise RuntimeError(msg)
        return confirmation

    @property
    def _connection(self) -> AbstractRobustConnection:
        if self._connection_ is None:
            msg = (
                f"{self} has no connection, please call"
                f" {self.__class__.__aenter__.__name__}"
            )
            raise ValueError(msg)
        return self._connection_

    @property
    def _channel(self) -> AbstractRobustChannel:
        if self._channel_ is None:
            msg = (
                f"{self} has no channel, please call"
                f" {self.__class__.__aenter__.__name__}"
            )
            raise ValueError(msg)
        return self._channel_

    @property
    def channel(self) -> AbstractRobustChannel:
        return self._channel

    @property
    def connection(self) -> AbstractRobustChannel:
        return self._connection

    @classmethod
    @lru_cache(maxsize=1)
    def default_task_routing(cls) -> Routing:
        return Routing(
            exchange=Exchange(name=AMQP_TASKS_X, type=ExchangeType.DIRECT),
            routing_key=AMQP_TASKS_ROUTING_KEY,
            queue_name=AMQP_TASKS_QUEUE,
            queue_args={"x-queue-type": "quorum"},
            dead_letter_routing=Routing(
                exchange=Exchange(name=AMQP_TASKS_DL_X, type=ExchangeType.DIRECT),
                routing_key=AMQP_TASKS_DL_ROUTING_KEY,
                queue_name=AMQP_TASKS_DL_QUEUE,
            ),
        )

    @classmethod
    @lru_cache(maxsize=1)
    def manager_evt_routing(cls) -> Routing:
        return Routing(
            exchange=Exchange(name=AMQP_MANAGER_EVENTS_X, type=ExchangeType.DIRECT),
            routing_key=AMQP_MANAGER_EVENTS_ROUTING_KEY,
            queue_name=AMQP_MANAGER_EVENTS_QUEUE,
            dead_letter_routing=Routing(
                exchange=Exchange(
                    name=AMQP_MANAGER_EVENTS_DL_X, type=ExchangeType.DIRECT
                ),
                routing_key=AMQP_MANAGER_EVENTS_DL_ROUTING_KEY,
                queue_name=AMQP_MANAGER_EVENTS_DL_QUEUE,
            ),
        )

    @classmethod
    @lru_cache(maxsize=1)
    def worker_evt_routing(cls) -> Routing:
        return Routing(
            exchange=Exchange(name=AMQP_WORKER_EVENTS_X, type=ExchangeType.FANOUT),
            routing_key=AMQP_WORKER_EVENTS_ROUTING_KEY,
            queue_name=AMQP_WORKER_EVENTS_QUEUE,
        )

    async def _get_queue_iterator(
        self,
        routing: Routing,
        *,
        declare_exchanges: bool,
        declare_queues: bool = True,
        durable_queues: bool = True,
    ) -> Tuple[AbstractQueueIterator, AbstractExchange, Optional[AbstractExchange]]:
        await self._exit_stack.enter_async_context(
            cast(AbstractAsyncContextManager, self._channel)
        )
        dlq_ex = None
        await self._create_routing(
            routing,
            declare_exchanges=declare_exchanges,
            declare_queues=declare_queues,
            durable_queues=durable_queues,
        )
        ex = await self._channel.get_exchange(routing.exchange.name, ensure=False)
        queue = await self._channel.get_queue(routing.queue_name, ensure=False)
        kwargs = dict()
        if self._inactive_after_s is not None:
            kwargs["timeout"] = self._inactive_after_s
        return queue.iterator(**kwargs), ex, dlq_ex

    async def _create_routing(
        self,
        routing: Routing,
        *,
        declare_exchanges: bool = True,
        declare_queues: bool = True,
        durable_queues: bool = True,
    ):
        if declare_exchanges:
            x = await self._channel.declare_exchange(
                routing.exchange.name, type=routing.exchange.type, durable=True
            )
        else:
            x = await self._channel.get_exchange(routing.exchange.name, ensure=True)
        queue_args = None
        if routing.queue_args is not None:
            queue_args = deepcopy(routing.queue_args)
        if routing.dead_letter_routing:
            await self._create_routing(
                routing.dead_letter_routing,
                declare_exchanges=declare_exchanges,
                declare_queues=declare_queues,
                durable_queues=durable_queues,
            )
            if queue_args is None:
                queue_args = dict()
            dlx_name = routing.dead_letter_routing.exchange.name
            dl_routing_key = routing.dead_letter_routing.routing_key
            # TODO: this could be passed through policy
            update = {
                "x-dead-letter-exchange": dlx_name,
                "x-dead-letter-routing-key": dl_routing_key,
            }
            queue_args.update(update)
        if declare_queues:
            if self._is_qpid:
                queue_args = dict()
            queue = await self._channel.declare_queue(
                routing.queue_name, durable=durable_queues, arguments=queue_args
            )
        else:
            queue = await self._channel.get_queue(routing.queue_name, ensure=True)
        await queue.bind(x, routing_key=routing.routing_key)


class AMQPManagementClient(AiohttpClient):
    def __init__(
        self,
        rabbitmq_management_url: str,
        *,
        rabbitmq_vhost: str,
        rabbitmq_auth: BasicAuth,
    ):
        super().__init__(rabbitmq_management_url, rabbitmq_auth)
        self._vhost = rabbitmq_vhost

    async def set_policy(self, policy: AMQPPolicy):
        url = f"/api/policies/{self._vhost}/{parse.quote(policy.name)}"
        data = {"pattern": policy.pattern, "definition": policy.definition}
        if policy.apply_to is not None:
            data["apply-to"] = policy.apply_to.value
        if policy.priority is not None:
            data["priority"] = policy.priority
        async with self._put(url, json=data):
            pass

    async def list_policies(self) -> List[Dict]:
        url = f"/api/policies/{self._vhost}"
        async with self._get(url) as res:
            return await res.json()

    async def clear_policies(self):
        policies = await self.list_policies()
        for p in policies:
            url = f"/api/policies/{self._vhost}/{p['name']}"
            async with self._delete(url):
                pass


def amqp_task_group_policy(
    task_routing: Routing,
    group: Optional[TaskGroup],
    app_max_task_queue_size: Optional[int],
) -> AMQPPolicy:
    pattern = rf"^{re.escape(task_routing.queue_name)}$"
    name = f"task-group-policy-{task_routing.queue_name}"
    definition = {"overflow": "reject-publish", "delivery-limit": 10}
    max_task_queue_size = app_max_task_queue_size
    if group is not None and group.max_task_queue_size is not None:
        max_task_queue_size = group.max_task_queue_size
    if max_task_queue_size is not None:
        definition["max-length"] = max_task_queue_size
    return AMQPPolicy(
        name=name,
        pattern=pattern,
        definition=definition,
        apply_to=ApplyTo.QUEUES,
        priority=AMQP_TASK_QUEUE_PRIORITY,
    )


class RobustChannel(RobustChannel_):
    async def __close_callback(self, _: Any, exc: BaseException) -> None:
        # pylint: disable=unused-private-member
        timeout_exc = parse_consumer_timeout(exc)
        if timeout_exc is not None:
            logger.error("channel closing due to consumer timeout: %s", exc)
            raise timeout_exc from exc


class RobustConnection(RobustConnection_):
    CHANNEL_CLASS: Type[RobustChannel] = RobustChannel

    # Defined async context manager attributes to be able to enter and exit this
    # in ExitStack
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def connect(self, timeout: TimeoutType = None) -> None:
        self.transport = await QpidUnderlayConnection.connect(
            self.url,
            self._on_connection_close,
            timeout=timeout,
            **self.kwargs,
        )
        await self._on_connected()


class QpidUnderlayConnection(UnderlayConnection):
    @classmethod
    async def make_connection(
        cls,
        url: URL,
        timeout: TimeoutType = None,
        **kwargs: Any,
    ) -> aiormq.abc.AbstractConnection:
        connection: aiormq.abc.AbstractConnection = await asyncio.wait_for(
            aiormq_qpid_connect(url, **kwargs), timeout=timeout
        )
        await connection.ready()
        return connection


async def aiormq_qpid_connect(
    url: URLorStr,
    *args: Any,
    client_properties: Optional[FieldTable] = None,
    **kwargs: Any,
) -> QpidConnection:
    connection = QpidConnection(url, *args, **kwargs)
    await connection.connect(client_properties or {})
    return connection


class QpidConnection(AiormqConnection):
    QPID_CAPABILITIES = {"publisher_confirms": False}

    @property
    def server_capabilities(self) -> ArgumentsType:
        # QPid doesn't seem to expose server properties so we mock empty ones
        return self.server_properties.get("capabilities", self.QPID_CAPABILITIES)


def parse_consumer_timeout(exc: BaseException) -> Optional[WorkerTimeoutError]:
    if not isinstance(exc, ChannelPreconditionFailed):
        return None
    if not exc.args:
        return None
    msg = exc.args[0]
    if _DELIVERY_ACK_TIMEOUT_RE.search(msg):
        return WorkerTimeoutError(msg)
    return None
